# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import logging
import socket
import sys
import time
import urllib

import pendulum
import rich_click as click
from duniterpy import constants as dp_const
from duniterpy.api.bma import blockchain
from duniterpy.api.client import Client
from duniterpy.documents.block import Block
from pydiscourse import DiscourseClient
from pydiscourse.exceptions import DiscourseClientError

from silkaj import constants
from silkaj.blockchain.tools import get_blockchain_parameters
from silkaj.network import client_instance
from silkaj.tools import get_currency_symbol
from silkaj.wot.tools import wot_lookup

G1_CESIUM_URL = "https://demo.cesium.app/"
GTEST_CESIUM_URL = "https://g1-test.cesium.app/"
CESIUM_BLOCK_PATH = "#/app/block/"

DUNITER_FORUM_URL = "https://forum.duniter.org/"
MONNAIE_LIBRE_FORUM_URL = "https://forum.monnaie-libre.fr/"

DUNITER_FORUM_G1_TOPIC_ID = 4393
DUNITER_FORUM_GTEST_TOPIC_ID = 6554
MONNAIE_LIBRE_FORUM_G1_TOPIC_ID = 30219  # 26117, 17627, 8233


@click.command(
    "exclusions",
    help="DeathReaper: Generate membership exclusions messages, \
markdown formatted and publish them on Discourse Forums",
)
@click.option(
    "-a",
    "--api-id",
    help="Username used on Discourse forum API",
)
@click.option(
    "-du",
    "--duniter-forum-api-key",
    help="API key used on Duniter Forum",
)
@click.option(
    "-ml",
    "--ml-forum-api-key",
    help="API key used on Monnaie Libre Forum",
)
@click.argument("days", default=1, type=click.FloatRange(0, 50))
@click.option(
    "--publish",
    is_flag=True,
    help="Publish the messages on the forums, otherwise it will be printed here.",
)
def exclusions_command(api_id, duniter_forum_api_key, ml_forum_api_key, days, publish):
    params = get_blockchain_parameters()
    currency = params["currency"]
    check_options(api_id, duniter_forum_api_key, ml_forum_api_key, publish, currency)
    bma_client = client_instance()
    blocks_to_process = get_blocks_to_process(bma_client, days, params)
    if not blocks_to_process:
        no_exclusion(days, currency)
    message = gen_message_over_blocks(bma_client, blocks_to_process, params)
    if not message:
        no_exclusion(days, currency)
    header = gen_header(blocks_to_process)
    # Add ability to publish just one of the two forum, via a flags?

    publish_display(
        api_id,
        duniter_forum_api_key,
        header + message,
        publish,
        currency,
        "duniter",
    )
    if currency == dp_const.G1_CURRENCY_CODENAME:
        publish_display(
            api_id,
            ml_forum_api_key,
            header + message,
            publish,
            currency,
            "monnaielibre",
        )


def check_options(api_id, duniter_forum_api_key, ml_forum_api_key, publish, currency):
    if publish and (
        not api_id
        or not duniter_forum_api_key
        or (not ml_forum_api_key and currency != dp_const.G1_TEST_CURRENCY_CODENAME)
    ):
        sys.exit(
            f"Error: To be able to publish, api_id, duniter_forum_api, and \
ml_forum_api_key (not required for {constants.GTEST_SYMBOL}) options should be specified",
        )


def no_exclusion(days, currency):
    # Use Humanize
    print(f"No exclusion to report within the last {days} day(s) on {currency}")
    # Success exit status for not failing GitLab job in case there is no exclusions
    sys.exit()


def get_blocks_to_process(bma_client, days, params):
    head_number = bma_client(blockchain.current)["number"]
    block_number_days_ago = (
        head_number - days * 24 * constants.ONE_HOUR / params["avgGenTime"]
    )
    # print(block_number_days_ago) # DEBUG

    i = 0
    blocks_with_excluded = bma_client(blockchain.excluded)["result"]["blocks"]
    for i, block_number in reversed(list(enumerate(blocks_with_excluded))):
        if block_number < block_number_days_ago:
            index = i
            break
    return blocks_with_excluded[index + 1 :]


def gen_message_over_blocks(bma_client, blocks_to_process, params):
    """
    Loop over the list of blocks to retrieve and parse the blocks
    Ignore revocation kind of exclusion
    """
    if params["currency"] == dp_const.G1_CURRENCY_CODENAME:
        es_client = Client(constants.G1_CSP_USER_ENDPOINT)
    else:
        es_client = Client(constants.GTEST_CSP_USER_ENDPOINT)
    message = ""
    for block_number in blocks_to_process:
        logging.info("Processing block number %s", block_number)
        print(f"Processing block number {block_number}")
        # DEBUG / to be removed once the #115 logging system is set

        try:
            block = bma_client(blockchain.block, block_number)
        except urllib.error.HTTPError:
            time.sleep(2)
            block = bma_client(blockchain.block, block_number)
        block_hash = block["hash"]
        block = Block.from_signed_raw(block["raw"] + block["signature"] + "\n")

        if block.revoked and block.excluded[0] == block.revoked[0].pubkey:
            continue
        message += generate_message(es_client, block, block_hash, params)
    return message


def gen_header(blocks_to_process):
    nbr_exclusions = len(blocks_to_process)
    # Handle when there is one block with multiple exclusion within
    # And when there is a revocation
    s = "s" if nbr_exclusions > 1 else ""
    des_du = "des" if nbr_exclusions > 1 else "du"
    currency_symbol = get_currency_symbol()
    header = f"## Exclusion{s} de la toile de confiance {currency_symbol}, perte{s} {des_du} statut{s} de membre"
    message_g1 = "\n> Message automatique. Merci de notifier vos proches de leur exclusion de la toile de confiance."
    return header + message_g1


def generate_message(es_client, block, block_hash, params):
    """
    Loop over exclusions within a block
    Generate identity header + info
    """
    message = ""
    for excluded in block.excluded:
        lookup = wot_lookup(excluded)[0]
        uid = lookup["uids"][0]["uid"]

        pubkey = lookup["pubkey"]
        try:
            response = es_client.get(f"user/profile/{pubkey}/_source")
            es_uid = response["title"]
        except (urllib.error.HTTPError, socket.timeout):
            es_uid = uid
            logging.info("Cesium+ API: Not found pubkey or connection error")

        if params["currency"] == dp_const.G1_CURRENCY_CODENAME:
            cesium_url = G1_CESIUM_URL
        else:
            cesium_url = GTEST_CESIUM_URL
        cesium_url += CESIUM_BLOCK_PATH
        message += f"\n\n### @{uid} [{es_uid}]({cesium_url}{block.number}/{block_hash}?ssl=true)\n"
        message += generate_identity_info(lookup, block, params)
    return message


def generate_identity_info(lookup, block, params):
    info = "- **Certifié·e par**"
    nbr_different_certifiers = 0
    for i, certifier in enumerate(lookup["uids"][0]["others"]):
        if certifier["uids"][0] not in info:
            nbr_different_certifiers += 1
        info += elements_inbetween_list(i, lookup["uids"][0]["others"])
        info += "@" + certifier["uids"][0]
    if lookup["signed"]:
        info += ".\n- **A certifié**"
    for i, certified in enumerate(lookup["signed"]):
        info += elements_inbetween_list(i, lookup["signed"])
        info += "@" + certified["uid"]
    dt = pendulum.from_timestamp(block.mediantime + constants.ONE_HOUR, tz="local")
    info += ".\n- **Exclu·e le** " + dt.format("LLLL zz", locale="fr")
    info += "\n- **Raison de l'exclusion** : "
    if nbr_different_certifiers < params["sigQty"]:
        info += "manque de certifications"
    else:
        info += "expiration du document d'adhésion"
    # a renouveller tous les ans (variable) humanize(params[""])
    return info


def elements_inbetween_list(i, cert_list):
    return " " if i == 0 else (" et " if i + 1 == len(cert_list) else ", ")


def publish_display(api_id, forum_api_key, message, publish, currency, forum):
    if publish:
        topic_id = get_topic_id(currency, forum)
        publish_message_on_the_forum(api_id, forum_api_key, message, topic_id, forum)
    elif forum == "duniter":
        click.echo(message)


def get_topic_id(currency, forum):
    if currency == dp_const.G1_CURRENCY_CODENAME:
        if forum == "duniter":
            return DUNITER_FORUM_G1_TOPIC_ID
        return MONNAIE_LIBRE_FORUM_G1_TOPIC_ID
    return DUNITER_FORUM_GTEST_TOPIC_ID


def publish_message_on_the_forum(api_id, forum_api_key, message, topic_id, forum):
    if forum == "duniter":
        discourse_client = DiscourseClient(
            DUNITER_FORUM_URL,
            api_username=api_id,
            api_key=forum_api_key,
        )
    else:
        discourse_client = DiscourseClient(
            MONNAIE_LIBRE_FORUM_URL,
            api_username=api_id,
            api_key=forum_api_key,
        )
    try:
        response = discourse_client.create_post(message, topic_id=topic_id)
        publication_link(forum, response, topic_id)
    except DiscourseClientError:
        logging.exception("Issue publishing on %s", forum)
        # Handle DiscourseClient exceptions, pass them to the logger

    # discourse_client.close()
    # How to close this client? It looks like it is not implemented
    # May be by closing requests' client


def publication_link(forum, response, topic_id):
    forum_url = DUNITER_FORUM_URL if forum == "duniter" else MONNAIE_LIBRE_FORUM_URL
    print(f"Published on {forum_url}t/{response['topic_slug']}/{topic_id!s}/last")



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import functools
import sys
from typing import Any, Union

import rich_click as click

from silkaj.blockchain.tools import get_blockchain_parameters
from silkaj.constants import FAILURE_EXIT_STATUS, G1_SYMBOL, GTEST_SYMBOL


@functools.lru_cache(maxsize=1)
def get_currency_symbol() -> str:
    params = get_blockchain_parameters()
    if params["currency"] == "g1":
        return G1_SYMBOL
    return GTEST_SYMBOL


@click.pass_context
def has_account_defined(
    ctx: click.Context,
    exit_error: bool = True,
) -> Union[bool, str]:
    if not (account_name := ctx.obj["ACCOUNT_NAME"]):
        if exit_error:
            click_fail("--account general option should be specified")
        return False
    return account_name


def message_exit(message: str) -> None:
    print(message)
    sys.exit(FAILURE_EXIT_STATUS)


@click.pass_context
def click_fail(context: click.Context, message: str) -> None:
    context.fail(message)


class MutuallyExclusiveOption(click.Option):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        _help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = (
                f"{_help} NOTE: This argument is mutually exclusive with arguments: [{ex_str}]."
            )
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx: click.Context, opts: Any, args: Any) -> Any:
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            arguments = ", ".join(self.mutually_exclusive)
            raise click.UsageError(
                message=f"Usage: `{self.name}` is mutually exclusive with arguments `{arguments}`.",
            )

        return super().handle_parse_result(ctx, opts, args)



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import re
from pathlib import Path
from typing import Optional

import rich_click as click
from duniterpy.key.scrypt_params import ScryptParams
from duniterpy.key.signing_key import SigningKey, SigningKeyException

from silkaj import tools
from silkaj.account_storage import AccountStorage
from silkaj.constants import PUBKEY_PATTERN
from silkaj.public_key import gen_pubkey_checksum

SEED_HEX_PATTERN = "^[0-9a-fA-F]{64}$"
PUBSEC_PUBKEY_PATTERN = f"pub: ({PUBKEY_PATTERN})"
PUBSEC_SIGNKEY_PATTERN = "sec: ([1-9A-HJ-NP-Za-km-z]{87,90})"


@click.pass_context
def auth_method(ctx: click.Context) -> SigningKey:
    """Account storage authentication"""
    password = ctx.obj["PASSWORD"]
    authfile = AccountStorage().authentication_file_path()
    wif_content = authfile.read_text()
    regex = re.compile("Type: ([a-zA-Z]+)", re.MULTILINE)
    match = re.search(regex, wif_content)
    if match and match.groups()[0] == "EWIF" and not password:
        password = click.prompt("Encrypted WIF, enter your password", hide_input=True)
    return auth_by_wif_file(authfile, password)


def auth_options(
    auth_file: Path,
    auth_seed: bool,
    auth_wif: bool,
    nrp: Optional[str] = None,
) -> SigningKey:
    """Authentication from CLI options"""
    if auth_file:
        return auth_by_auth_file(auth_file)
    if auth_seed:
        return auth_by_seed()
    if auth_wif:
        return auth_by_wif()
    return auth_by_scrypt(nrp)


@click.command("authentication", help="Generate and store authentication file")
@click.option(
    "--auth-scrypt",
    "--scrypt",
    is_flag=True,
    help="Scrypt authentication. Default method",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["auth_file", "auth_seed", "auth_wif"],
)
@click.option("--nrp", help='Scrypt parameters: defaults N,r,p: "4096,16,1"')
@click.option(
    "--auth-file",
    "-af",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Seed hexadecimal authentication from file path",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["auth_scrypt", "auth_seed", "auth_wif"],
)
@click.option(
    "--auth-seed",
    "--seed",
    is_flag=True,
    help="Seed hexadecimal authentication",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["auth_scrypt", "auth_file", "auth_wif"],
)
@click.option(
    "--auth-wif",
    "--wif",
    is_flag=True,
    help="WIF and EWIF authentication methods",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["auth_scrypt", "auth_file", "auth_seed"],
)
@click.option(
    "--password",
    "-p",
    help="EWIF encryption password for the destination file. \
If no password argument is passed, WIF format will be used. \
If you use this option prefix the command \
with a space so the password does not get saved in your shell history. \
Password input will be suggested via a prompt.",
)
@click.pass_context
def generate_auth_file(
    ctx: click.Context,
    auth_scrypt: bool,
    nrp: Optional[str],
    auth_file: Path,
    auth_seed: bool,
    auth_wif: bool,
    password: Optional[str],
) -> None:
    auth_file_path = AccountStorage().authentication_file_path(check_exist=False)

    if not password and click.confirm(
        "Would you like to encrypt the generated authentication file?",
    ):
        password = click.prompt("Enter encryption password", hide_input=True)

    if password:
        password_confirmation = click.prompt(
            "Enter encryption password confirmation",
            hide_input=True,
        )
        if password != password_confirmation:
            tools.click_fail("Entered passwords differ")

    key = auth_options(auth_file, auth_seed, auth_wif, nrp)
    pubkey_cksum = gen_pubkey_checksum(key.pubkey)
    if auth_file_path.is_file():
        message = (
            f"Would you like to erase {auth_file_path} with an authentication file corresponding \
to following pubkey `{pubkey_cksum}`?"
        )
        click.confirm(message, abort=True)
    if password:
        key.save_ewif_file(auth_file_path, password)
    else:
        key.save_wif_file(auth_file_path)
    print(
        f"Authentication file '{auth_file_path}' generated and stored for public key: {pubkey_cksum}",
    )


@click.pass_context
def auth_by_auth_file(ctx: click.Context, authfile: Path) -> SigningKey:
    """
    Uses an authentication file to generate the key
    Authfile can either be:
    * A seed in hexadecimal encoding
    * PubSec format with public and private key in base58 encoding
    """
    filetxt = authfile.read_text(encoding="utf-8")

    # two regural expressions for the PubSec format
    regex_pubkey = re.compile(PUBSEC_PUBKEY_PATTERN, re.MULTILINE)
    regex_signkey = re.compile(PUBSEC_SIGNKEY_PATTERN, re.MULTILINE)

    # Seed hexadecimal format
    if re.search(re.compile(SEED_HEX_PATTERN), filetxt):
        return SigningKey.from_seedhex_file(authfile)
    # PubSec format
    if re.search(regex_pubkey, filetxt) and re.search(regex_signkey, filetxt):
        return SigningKey.from_pubsec_file(authfile)
    tools.click_fail("The format of the file is invalid")
    return None


def auth_by_seed() -> SigningKey:
    seedhex = click.prompt("Please enter your seed on hex format", hide_input=True)
    try:
        return SigningKey.from_seedhex(seedhex)
    except SigningKeyException as error:
        tools.click_fail(error)


@click.pass_context
def auth_by_scrypt(ctx: click.Context, nrp: Optional[str]) -> SigningKey:
    salt = click.prompt(
        "Please enter your Scrypt Salt (Secret identifier)",
        hide_input=True,
        default="",
    )
    password = click.prompt(
        "Please enter your Scrypt password (masked)",
        hide_input=True,
        default="",
    )

    if nrp:
        a, b, c = nrp.split(",")

        if a.isnumeric() and b.isnumeric() and c.isnumeric():
            n, r, p = int(a), int(b), int(c)
            if n <= 0 or n > 65536 or r <= 0 or r > 512 or p <= 0 or p > 32:
                tools.click_fail("The values of Scrypt parameters are not good")
            scrypt_params = ScryptParams(n, r, p)
        else:
            tools.click_fail("one of n, r or p is not a number")
    else:
        scrypt_params = None

    try:
        return SigningKey.from_credentials(salt, password, scrypt_params)
    except SigningKeyException as error:
        tools.click_fail(error)


def auth_by_wif() -> SigningKey:
    wif_hex = click.prompt(
        "Enter your WIF or Encrypted WIF address (masked)",
        hide_input=True,
    )
    password = click.prompt(
        "(Leave empty in case WIF format) Enter the Encrypted WIF password (masked)",
        hide_input=True,
    )
    try:
        return SigningKey.from_wif_or_ewif_hex(wif_hex, password)
    except SigningKeyException as error:
        tools.click_fail(error)


def auth_by_wif_file(wif_file: Path, password: Optional[str] = None) -> SigningKey:
    try:
        return SigningKey.from_wif_or_ewif_file(wif_file, password)
    except SigningKeyException as error:
        tools.click_fail(error)



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import csv
from operator import eq, itemgetter, ne, neg
from pathlib import Path
from typing import Any, Optional
from urllib.error import HTTPError

import pendulum
import rich_click as click
from duniterpy.api.bma.tx import history
from duniterpy.api.client import Client
from duniterpy.documents.transaction import OutputSource, Transaction
from duniterpy.grammars.output import Condition

from silkaj.constants import ALL, ALL_DIGITAL, CENT_MULT_TO_UNIT
from silkaj.money import tools as mt
from silkaj.network import client_instance
from silkaj.public_key import (
    check_pubkey_format,
    gen_pubkey_checksum,
    validate_checksum,
)
from silkaj.tools import get_currency_symbol
from silkaj.tui import Table
from silkaj.wot import tools as wt


@click.command("history", help="History of wallet money movements")
@click.argument("pubkey")
@click.option("--uids", "-u", is_flag=True, help="Display identities username")
@click.option(
    "--full-pubkey", "-f", is_flag=True, help="Display full-length public keys"
)
@click.option(
    "--csv-file",
    "--csv",
    type=click.Path(exists=False, writable=True, dir_okay=False, path_type=Path),
    help="Write in specified file name in CSV (Comma-separated values) format the history of money movements",
)
def transaction_history(
    pubkey: str,
    uids: bool,
    full_pubkey: bool,
    csv_file: Optional[Path],
) -> None:
    if csv_file:
        full_pubkey = True

    if check_pubkey_format(pubkey):
        pubkey = validate_checksum(pubkey)

    client = client_instance()
    ud_value = mt.get_ud_value()
    currency_symbol = get_currency_symbol()

    received_txs, sent_txs = [], []  # type: list[Transaction], list[Transaction]
    get_transactions_history(client, pubkey, received_txs, sent_txs)
    remove_duplicate_txs(received_txs, sent_txs)

    txs_list = generate_txs_list(
        received_txs,
        sent_txs,
        pubkey,
        ud_value,
        currency_symbol,
        uids,
        full_pubkey,
    )
    table_headers = [
        "Date",
        "Issuers/Recipients",
        f"Amounts {currency_symbol}",
        f"Amounts UD{currency_symbol}",
        "Reference",
    ]
    if csv_file:
        if csv_file.is_file():
            click.confirm(f"{csv_file} exists, would you like to erase it?", abort=True)
        txs_list.insert(0, table_headers)
        with csv_file.open("w", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerows(txs_list)
        click.echo(f"{csv_file} file successfully saved!")
    else:
        table = Table()
        table.fill_rows(txs_list, table_headers)
        header = generate_header(pubkey, currency_symbol, ud_value)
        click.echo_via_pager(header + table.draw())


def generate_header(pubkey: str, currency_symbol: str, ud_value: int) -> str:
    try:
        idty = wt.identity_of(pubkey)
    except HTTPError:
        idty = {"uid": ""}
    balance = mt.get_amount_from_pubkey(pubkey)
    balance_ud = round(balance[1] / ud_value, 2)
    date = pendulum.now().format(ALL)
    return f"Transactions history from: {idty['uid']} {gen_pubkey_checksum(pubkey)}\n\
Current balance: {balance[1] / CENT_MULT_TO_UNIT} {currency_symbol}, {balance_ud} UD {currency_symbol} on {date}\n"


def get_transactions_history(
    client: Client,
    pubkey: str,
    received_txs: list,
    sent_txs: list,
) -> None:
    """
    Get transaction history
    Store txs in Transaction object
    """
    tx_history = client(history, pubkey)
    currency = tx_history["currency"]

    for received in tx_history["history"]["received"]:
        received_txs.append(Transaction.from_bma_history(received, currency))
    for sent in tx_history["history"]["sent"]:
        sent_txs.append(Transaction.from_bma_history(sent, currency))


def remove_duplicate_txs(received_txs: list, sent_txs: list) -> None:
    """
    Remove duplicate transactions from history
    Remove received tx which contains output back return
    that we don't want to displayed
    A copy of received_txs is necessary to remove elements
    """
    for received_tx in list(received_txs):
        if received_tx in sent_txs:
            received_txs.remove(received_tx)


def generate_txs_list(
    received_txs: list[Transaction],
    sent_txs: list[Transaction],
    pubkey: str,
    ud_value: int,
    currency_symbol: str,
    uids: bool,
    full_pubkey: bool,
) -> list:
    """
    Generate information in a list of lists for texttable
    Merge received and sent txs
    Sort txs temporarily
    """

    received_txs_list, sent_txs_list = (
        [],
        [],
    )  # type: list[Transaction], list[Transaction]
    parse_received_tx(
        received_txs_list,
        received_txs,
        pubkey,
        ud_value,
        uids,
        full_pubkey,
    )
    parse_sent_tx(sent_txs_list, sent_txs, pubkey, ud_value, uids, full_pubkey)
    txs_list = received_txs_list + sent_txs_list

    txs_list.sort(key=itemgetter(0), reverse=True)
    return txs_list


def parse_received_tx(
    received_txs_table: list[Transaction],
    received_txs: list[Transaction],
    pubkey: str,
    ud_value: int,
    uids: bool,
    full_pubkey: bool,
) -> None:
    """
    Extract issuers` pubkeys
    Get identities from pubkeys
    Convert time into human format
    Assign identities
    Get amounts and assign amounts and amounts_ud
    Append reference/comment
    """
    issuers = []
    for received_tx in received_txs:
        for issuer in received_tx.issuers:
            issuers.append(issuer)
    identities = wt.identities_from_pubkeys(issuers, uids)
    for received_tx in received_txs:
        for issuer in received_tx.issuers:
            tx_list = []
            tx_list.append(
                pendulum.from_timestamp(received_tx.time, tz="local").format(
                    ALL_DIGITAL
                )
            )
            tx_list.append(assign_idty_from_pubkey(issuer, identities, full_pubkey))
            amounts = tx_amount(received_tx, pubkey, received_func)[0]
            tx_list.append(amounts / CENT_MULT_TO_UNIT)
            tx_list.append(round(amounts / ud_value, 2))
            tx_list.append(received_tx.comment)
            received_txs_table.append(tx_list)


def parse_sent_tx(
    sent_txs_table: list[Transaction],
    sent_txs: list[Transaction],
    pubkey: str,
    ud_value: int,
    uids: bool,
    full_pubkey: bool,
) -> None:
    """
    Extract recipients` pubkeys from outputs
    Get identities from pubkeys
    Convert time into human format
    If not output back return:
    Assign amounts, amounts_ud, identities, and comment
    """
    pubkeys = []
    for sent_tx in sent_txs:
        outputs = tx_amount(sent_tx, pubkey, sent_func)[1]
        for output in outputs:
            if output_available(output.condition, ne, pubkey):
                pubkeys.append(output.condition.left.pubkey)

    identities = wt.identities_from_pubkeys(pubkeys, uids)
    for sent_tx in sent_txs:
        total_amount, outputs = tx_amount(sent_tx, pubkey, sent_func)
        for output in outputs:
            if output_available(output.condition, ne, pubkey):
                tx_list = []
                tx_list.append(
                    pendulum.from_timestamp(sent_tx.time, tz="local").format(
                        ALL_DIGITAL
                    )
                )
                wallet = assign_idty_from_pubkey(
                    output.condition.left.pubkey, identities, full_pubkey
                )
                tx_list.append(wallet)
                tx_list.append(
                    neg(mt.amount_in_current_base(output)) / CENT_MULT_TO_UNIT,
                )
                tx_list.append(
                    round(neg(mt.amount_in_current_base(output)) / ud_value, 2),
                )
                tx_list.append(sent_tx.comment)
                sent_txs_table.append(tx_list)


def tx_amount(
    tx: list[Transaction],
    pubkey: str,
    function: Any,
) -> tuple[int, list[OutputSource]]:
    """
    Determine transaction amount from output sources
    """
    amount = 0
    outputs = []
    for output in tx.outputs:  # type: ignore[attr-defined]
        if output_available(output.condition, ne, pubkey):
            outputs.append(output)
        amount += function(output, pubkey)
    return amount, outputs


def received_func(output: OutputSource, pubkey: str) -> int:
    if output_available(output.condition, eq, pubkey):
        return mt.amount_in_current_base(output)
    return 0


def sent_func(output: OutputSource, pubkey: str) -> int:
    if output_available(output.condition, ne, pubkey):
        return neg(mt.amount_in_current_base(output))
    return 0


def output_available(condition: Condition, comparison: Any, value: str) -> bool:
    """
    Check if output source is available
    Currently only handle simple SIG condition
    XHX, CLTV, CSV should be handled when present in the blockchain
    """
    if hasattr(condition.left, "pubkey"):
        return comparison(condition.left.pubkey, value)
    return False


def assign_idty_from_pubkey(pubkey: str, identities: list, full_pubkey: bool) -> str:
    idty = gen_pubkey_checksum(pubkey, short=not full_pubkey)
    for identity in identities:
        if pubkey == identity["pubkey"]:
            pubkey_mod = gen_pubkey_checksum(pubkey, short=not full_pubkey)
            idty = f"{identity['uid']} - {pubkey_mod}"
    return idty



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import shutil
import sys
import urllib
from typing import Union

import pendulum
import rich_click as click
from duniterpy.api import bma
from duniterpy.documents import BlockID, Identity, Revocation
from texttable import Texttable

from silkaj.constants import ALL
from silkaj.network import client_instance
from silkaj.public_key import gen_pubkey_checksum
from silkaj.wot.tools import wot_lookup


def display_identity(idty: Identity) -> Texttable:
    """
    Creates a table containing the identity infos
    """
    client = client_instance()
    id_table = []
    id_table.append(["Public key", gen_pubkey_checksum(idty.pubkey)])
    id_table.append(["User ID", idty.uid])
    id_table.append(["Blockstamp", str(idty.block_id)])
    creation_block = client(bma.blockchain.block, idty.block_id.number)
    creation_date = pendulum.from_timestamp(creation_block["time"], tz="local").format(
        ALL,
    )
    id_table.append(["Created on", creation_date])
    # display infos
    table = Texttable(max_width=shutil.get_terminal_size().columns)
    table.add_rows(id_table, header=False)
    return table


def check_many_identities(document: Union[Identity, Revocation]) -> bool:
    """
    Checks if many identities match the one looked after.
    Returns True if the same identity is found, False if not.
    """
    doc_type = document.__class__.__name__
    error_no_identical_id = f"{doc_type} document does not match any valid identity."
    idty = document if doc_type == "Identity" else document.identity

    try:
        results_pubkey = wot_lookup(idty.pubkey)
        results_uid = wot_lookup(idty.uid)
    except urllib.error.HTTPError:
        sys.exit(
            f"{error_no_identical_id}\nuid: {idty.uid}\npubkey: \
{gen_pubkey_checksum(idty.pubkey)}",
        )

    # get all matching identities
    lookup_ids = merge_ids_lists(results_pubkey, results_uid, idty.currency)
    match = False
    for n, lookup in enumerate(lookup_ids):
        if idty == lookup:
            lookup_ids.pop(n)
            match = True
            break
    alternate_ids = display_alternate_ids(lookup_ids).draw()
    if match:
        if len(lookup_ids) >= 1:
            click.echo(f"One matching identity!\nSimilar identities:\n{alternate_ids}")
        return True
    click.echo(f"{error_no_identical_id}\nSimilar identities:\n{alternate_ids}")
    return False


def display_alternate_ids(ids_list: list) -> Texttable:
    labels = ["uid", "public key", "timestamp"]
    table = Texttable(max_width=shutil.get_terminal_size().columns)
    table.header(labels)
    for _id in ids_list:
        table.add_row(
            [_id.uid, gen_pubkey_checksum(_id.pubkey), str(_id.block_id)[:12]],
        )
    return table


def merge_ids_lists(lookups_pubkey: list, lookups_uid: list, currency: str) -> list:
    """
    merge two lists of identities and remove duplicate identities.
    """
    ids = ids_list_from_lookups(lookups_pubkey, currency)
    ids_uid = ids_list_from_lookups(lookups_uid, currency)
    for _id in ids_uid:
        # __equal__ does not work. This is condition "id in ids".
        for listed_id in ids:
            if _id.signed_raw() == listed_id.signed_raw():
                id_in_ids = True
                break
            id_in_ids = False
        if not id_in_ids:
            ids.append(_id)
    return ids


def ids_list_from_lookups(lookups: list, currency: str) -> list:
    ids = []
    for lookup in lookups:
        pubkey = lookup["pubkey"]
        lookup_ids = lookup["uids"]
        for _id in lookup_ids:
            appended_id = Identity(
                currency=currency,
                pubkey=pubkey,
                uid=_id["uid"],
                block_id=BlockID.from_str(_id["meta"]["timestamp"]),
            )
            appended_id.signature = _id["self"]
            ids.append(appended_id)
    return ids



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

from operator import itemgetter
from os import system

import jsonschema
import pendulum
import rich_click as click
from duniterpy.api import bma
from duniterpy.api.client import WSConnection
from websocket._exceptions import WebSocketConnectionClosedException

from silkaj import network, tui
from silkaj.constants import ALL


@click.command(
    "difficulty",
    help="Display current Proof-of-Work difficulty level to generate next block",
)
def difficulties() -> None:
    client = network.client_instance()
    try:
        ws = client(bma.ws.block)
        while True:
            current = ws.receive_json()
            jsonschema.validate(current, bma.ws.WS_BLOCK_SCHEMA)
            diffi = client(bma.blockchain.difficulties)
            display_diffi(current, diffi)
    except (jsonschema.ValidationError, WebSocketConnectionClosedException) as e:
        print(f"{e.__class__.__name__!s}: {e!s}")


def display_diffi(current: WSConnection, diffi: dict) -> None:
    issuers = 0
    sorted_diffi = sorted(diffi["levels"], key=itemgetter("level"), reverse=True)
    for d in diffi["levels"]:
        if d["level"] / 2 < current["powMin"]:
            issuers += 1
        d["match"] = match_pattern(d["level"])[0][:20]
        d["Π diffi"] = compute_power(match_pattern(d["level"])[1])
        d["Σ diffi"] = d.pop("level")
    system("cls||clear")
    block_gen = pendulum.from_timestamp(current["time"], tz="local").format(ALL)
    match = match_pattern(int(current["powMin"]))[0]

    table = tui.Table(style="columns").set_cols_dtype(["t", "t", "t", "i"])
    table.fill_from_dict_list(sorted_diffi)

    content = f"Current block: n°{current['number']}, generated on {block_gen}\n\
Generation of next block n°{diffi['block']} \
possible by at least {issuers}/{len(diffi['levels'])} members\n\
Common Proof-of-Work difficulty level: {current['powMin']}, hash starting with `{match}`\n\
{table.draw()}"
    print(content)


def match_pattern(_pow: int, match: str = "", p: int = 1) -> tuple[str, int]:
    while _pow > 0:
        if _pow >= 16:
            match += "0"
            _pow -= 16
            p *= 16
        else:
            match += f"[0-{hex(15 - _pow)[2:].upper()}]"
            p *= _pow
            _pow = 0
    return f"{match}*", p


def compute_power(nbr: float, power: int = 0) -> str:
    while nbr >= 10:
        nbr /= 10
        power += 1
    return f"{nbr:.1f} x 10^{power}"



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import contextlib

import rich_click as click
from duniterpy.api.endpoint import endpoint as du_endpoint

from silkaj import tools
from silkaj.about import about
from silkaj.auth import generate_auth_file
from silkaj.blockchain.blocks import list_blocks
from silkaj.blockchain.difficulty import difficulties
from silkaj.blockchain.information import currency_info
from silkaj.checksum import checksum_command
from silkaj.constants import (
    G1_DEFAULT_ENDPOINT,
    G1_TEST_DEFAULT_ENDPOINT,
    SILKAJ_VERSION,
)
from silkaj.g1_monetary_license import license_command
from silkaj.money.balance import balance_cmd
from silkaj.money.history import transaction_history
from silkaj.money.transfer import transfer_money
from silkaj.wot import certify, revocation
from silkaj.wot.lookup import lookup_cmd
from silkaj.wot.membership import send_membership
from silkaj.wot.status import status

click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.OPTION_GROUPS = {
    "silkaj": [
        {
            "name": "Basic options",
            "options": ["--help", "--version"],
        },
        {
            "name": "Endpoint and currency specification",
            "options": ["--endpoint", "--gtest"],
        },
        {
            "name": "Account and authentication specification",
            "options": ["--account", "--password"],
        },
    ],
}


@click.group()
@click.help_option("-h", "--help")
@click.version_option(SILKAJ_VERSION, "-v", "--version")
@click.option(
    "--endpoint",
    "-ep",
    help=f"Without specifying this option, the default endpoint reaches Ğ1 currency on its official endpoint: https://{du_endpoint(G1_DEFAULT_ENDPOINT).host}. \
--endpoint allows to specify a custom endpoint following `<host>:<port>/<path>` format. \
`port` and `path` are optional. In case no port is specified, it defaults to 443.",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["gtest"],
)
@click.option(
    "--gtest",
    "-gt",
    is_flag=True,
    help=f"Uses official ĞTest currency endpoint: https://{du_endpoint(G1_TEST_DEFAULT_ENDPOINT).host}",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["endpoint"],
)
@click.option(
    "account_name",
    "--account",
    "-a",
    help="Account name used in storage `$HOME/.local/share/silkaj/$currency/$account_name` for authentication and revocation.",
)
@click.option(
    "--password",
    "-p",
    help="EWIF authentication password. If you use this option, prefix the command \
with a space so the password is not saved in your shell history. \
In case of an encrypted file, password input will be prompted.",
)
@click.option(
    "--display",
    "-d",
    is_flag=True,
    help="Display the generated document before sending it",
)
@click.option(
    "--dry-run",
    "-n",
    is_flag=True,
    help="By-pass the licence and confirmation. Do not send the document, but display it instead",
)
@click.pass_context
def cli(
    ctx: click.Context,
    endpoint: str,
    gtest: bool,
    account_name: str,
    password: str,
    display: bool,
    dry_run: bool,
) -> None:
    if display and dry_run:
        ctx.fail("Display and dry-run options can not be used together")

    ctx.obj = {}
    ctx.ensure_object(dict)
    ctx.obj["ENDPOINT"] = endpoint
    ctx.obj["GTEST"] = gtest
    ctx.obj["ACCOUNT_NAME"] = account_name
    ctx.obj["PASSWORD"] = password
    ctx.obj["DISPLAY_DOCUMENT"] = display
    ctx.obj["DRY_RUN"] = dry_run
    ctx.help_option_names = ["-h", "--help"]


cli.add_command(about)
cli.add_command(generate_auth_file)
cli.add_command(checksum_command)
cli.add_command(license_command)


@cli.group("blockchain", help="Blockchain related commands")
def blockchain_group() -> None:
    pass


blockchain_group.add_command(list_blocks)
blockchain_group.add_command(difficulties)
blockchain_group.add_command(currency_info)


@cli.group("money", help="Money management related commands")
def money_group() -> None:
    pass


money_group.add_command(balance_cmd)
money_group.add_command(transaction_history)
money_group.add_command(transfer_money)


@cli.group("wot", help="Web-of-Trust related commands")
def wot_group() -> None:
    pass


wot_group.add_command(certify.certify)
with contextlib.suppress(ModuleNotFoundError):
    from silkaj.wot.exclusions import exclusions_command

    wot_group.add_command(exclusions_command)

wot_group.add_command(lookup_cmd)
wot_group.add_command(send_membership)
wot_group.add_command(status)


@wot_group.group("revocation", help="Manage revocation document commands.")
def revocation_group() -> None:
    pass


revocation_group.add_command(revocation.create)
revocation_group.add_command(revocation.verify)
revocation_group.add_command(revocation.publish)
revocation_group.add_command(revocation.revoke_now)



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import pendulum
import rich_click as click
from duniterpy.api.bma import blockchain, wot

from silkaj.blockchain.tools import get_blockchain_parameters
from silkaj.constants import DATE
from silkaj.network import client_instance
from silkaj.public_key import gen_pubkey_checksum, is_pubkey_and_check
from silkaj.tui import Table
from silkaj.wot import tools as wt


@click.command(
    "status",
    help="Check received and sent certifications and \
consult the membership status of any given identity",
)
@click.argument("uid_pubkey")
def status(uid_pubkey: str) -> None:
    """
    get searched id
    get id of received and sent certifications
    display in a table the result with the numbers
    """
    client = client_instance()
    first_block = client(blockchain.block, 1)
    time_first_block = first_block["time"]

    checked_pubkey = is_pubkey_and_check(uid_pubkey)
    if checked_pubkey:
        uid_pubkey = str(checked_pubkey)

    identity, pubkey, signed = wt.choose_identity(uid_pubkey)
    certifications = {}  # type: dict
    params = get_blockchain_parameters()

    req = (client(wot.requirements, search=pubkey, pubkey=True))["identities"][0]
    certifications["received_expire"] = []
    certifications["received"] = []
    certifications["sent"] = []
    certifications["sent_expire"] = []
    for lookup_cert in identity["others"]:
        for req_cert in req["certifications"]:
            if req_cert["from"] == lookup_cert["pubkey"]:
                certifications["received_expire"].append(
                    pendulum.now().add(seconds=req_cert["expiresIn"]).format(DATE),
                )
                certifications["received"].append(f"{lookup_cert['uids'][0]} ✔")
                break
    for pending_cert in req["pendingCerts"]:
        certifications["received"].append(
            f"{(wt.identity_of(pending_cert['from']))['uid']} ✘",
        )
        certifications["received_expire"].append(
            pendulum.from_timestamp(pending_cert["expires_on"], tz="local").format(
                DATE
            ),
        )
    certifications["sent"], certifications["sent_expire"] = get_sent_certifications(
        signed,
        time_first_block,
        params,
    )

    nbr_sent_certs = len(certifications["sent"]) if "sent" in certifications else 0

    table = Table(style="columns").set_cols_align(["r", "r", "r", "r"])
    table.fill_from_dict(certifications)

    print(
        f"{identity['uid']} ({gen_pubkey_checksum(pubkey, True)}) \
from block #{identity['meta']['timestamp'][:15]}…\n\
received {len(certifications['received'])} and sent \
{nbr_sent_certs}/{params['sigStock']} certifications:\n\
{table.draw()}\n\
✔: Certification written in the blockchain\n\
✘: Pending certification, deadline treatment\n",
    )
    membership_status(certifications, pubkey, req)


def membership_status(certifications: dict, pubkey: str, req: dict) -> None:
    params = get_blockchain_parameters()
    if len(certifications["received"]) >= params["sigQty"]:
        date = certifications["received_expire"][
            len(certifications["received"]) - params["sigQty"]
        ]
        print(f"Membership expiration due to certification expirations: {date}")
    member_lookup = wt.is_member(pubkey)
    is_member = bool(member_lookup)
    print("member:", is_member)
    if req["revoked"]:
        revoke_date = pendulum.from_timestamp(req["revoked_on"], tz="local").format(
            DATE
        )
        print(f"revoked: {req['revoked']}\nrevoked on: {revoke_date}")
    if not is_member and req["wasMember"]:
        print("expired:", req["expired"], "\nwasMember:", req["wasMember"])
    elif is_member:
        expiration_date = (
            pendulum.now().add(seconds=req["membershipExpiresIn"]).format(DATE)
        )
        print(f"Membership document expiration: {expiration_date}")
        print("Sentry:", req["isSentry"])
    print("outdistanced:", req["outdistanced"])


def get_sent_certifications(
    signed: list,
    time_first_block: int,
    params: dict,
) -> tuple[list[str], list[str]]:
    sent = []
    expire = []
    if signed:
        for cert in signed:
            sent.append(cert["uid"])
            expire.append(
                expiration_date_from_block_id(
                    cert["cert_time"]["block"],
                    time_first_block,
                    params,
                ),
            )
    return sent, expire


def expiration_date_from_block_id(
    block_id: str,
    time_first_block: int,
    params: dict,
) -> str:
    expir_timestamp = (
        date_approximation(block_id, time_first_block, params["avgGenTime"])
        + params["sigValidity"]
    )
    return pendulum.from_timestamp(expir_timestamp, tz="local").format(DATE)


def date_approximation(block_id, time_first_block, avgentime):
    return time_first_block + block_id * avgentime



# Copyright  2016-2025 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path

from silkaj import tools
from silkaj.blockchain import tools as bc_tools


class AccountStorage:
    xdg_data_home = ".local/share"
    program_name = "silkaj"
    revocation_file_name = "revocation.txt"
    authentication_v1_file_name = "authentication_file_ed25519.dewif"
    authentication_v2_file_name = "authentication_file_sr25519.json"

    def __init__(self) -> None:
        self.account_name = tools.has_account_defined()

        self.path = Path.home().joinpath(
            self.xdg_data_home,
            self.program_name,
            bc_tools.get_currency(),
            self.account_name,
        )
        self.path.mkdir(parents=True, exist_ok=True)

    def authentication_file_path(self, check_exist: bool = True) -> Path:
        auth_file_path = self.path.joinpath(self.authentication_v1_file_name)
        if check_exist and not auth_file_path.is_file():
            tools.click_fail(
                f"{auth_file_path} not found for account name: {self.account_name}",
            )
        return auth_file_path

    def revocation_path(self, check_exist: bool = True) -> Path:
        revocation_path = self.path.joinpath(self.revocation_file_name)
        if check_exist and not revocation_path.is_file():
            tools.click_fail(
                f"{revocation_path} not found for account name: {self.account_name}",
            )
        return revocation_path
